#include <linux/proc_fs.h>
#include <asm/uaccess.h>

#include <linux/seq_file.h>
#include <mach/mt_spm.h>
#include <mach/mt_idle.h>

extern SPM_PCM_CONFIG pcm_config_suspend;
extern SPM_PCM_CONFIG pcm_config_dpidle;
extern SPM_PCM_CONFIG pcm_config_mcdi;
extern const char *pcm_scenario[SPM_PCM_SCENARIO_NUM];
static SPM_PCM_CONFIG *pcm_fs_curr = &pcm_config_suspend;

bool spm_fs_initialized = 0;

u32 pcm_wakesrc_backup[SPM_PCM_SCENARIO_NUM] = { 0, 0, 0 };

/*************************************************Common Part by Scott***************************************************************/

static int pcm_sysfs_read(struct seq_file *m, void *v)
{
	if (pcm_fs_curr->scenario == SPM_PCM_KERNEL_SUSPEND)
		seq_puts(m, "pcm_config_suspend={\n");
	else if (pcm_fs_curr->scenario == SPM_PCM_DEEP_IDLE)
		seq_puts(m, "pcm_config_dpidle={\n");
	else if (pcm_fs_curr->scenario == SPM_PCM_MCDI)
		seq_puts(m, "pcm_config_mcdi={\n");
	else {
		seq_puts(m, "SCENARIO ERROR\n");
		goto error;
	}

	seq_printf(m, "     .scenario = %d\n", pcm_fs_curr->scenario);
	seq_printf(m, "     .spm_turn_off_26m = %d\n", pcm_fs_curr->spm_turn_off_26m);
	seq_printf(m, "     .pcm_pwrlevel = %d\n", pcm_fs_curr->pcm_pwrlevel);
	seq_printf(m, "     .pcm_firmware_addr = 0x%X\n", pcm_fs_curr->pcm_firmware_addr);
	seq_printf(m, "     .pcm_firmware_len = %d\n", pcm_fs_curr->pcm_firmware_len);
	seq_printf(m, "     .spm_request_uart_sleep = %d\n", pcm_fs_curr->spm_request_uart_sleep);
	seq_printf(m, "     .md_mask = 0x%X  /*{MDCONN_MASK=0,CONN_MASK,MD_MASK,MDCONN_UNMASK}*/\n",
		   pcm_fs_curr->md_mask);
	seq_printf(m, "     .mm_mask = 0x%X  /*{MMALL_MASK=0,MFG_MASK,DISP_MASK,MMALL_UNMASK}*/\n",
		   pcm_fs_curr->mm_mask);
	seq_printf(m, "     .wfi_scu_mask = %d\n ", pcm_fs_curr->wfi_scu_mask);
	seq_printf(m, "     .wfi_l2c_mask = %d\n ", pcm_fs_curr->wfi_l2c_mask);
	seq_printf(m, "     .wfi_op = %d    /*{REDUCE_OR=0,REDUCE_AND}*/\n", pcm_fs_curr->wfi_op);
	seq_printf(m, "     .wfi_sel[]={%d,%d}\n", pcm_fs_curr->wfi_sel[0],
		   pcm_fs_curr->wfi_sel[1]);
	seq_printf(m, "     .timer_val_ms=%d\n", (u32) pcm_fs_curr->timer_val_ms);
	seq_printf(m, "     .wake_src= 0x%X\n", pcm_fs_curr->wake_src);
	seq_printf(m, "     .cpu_pdn= %d\n", pcm_fs_curr->cpu_pdn);
	seq_printf(m, "     .infra_pdn = %d\n", pcm_fs_curr->infra_pdn);
	seq_printf(m, "     .reserved = 0x%X\n", pcm_fs_curr->reserved);

	seq_puts(m, "}\n");

error:
	return 0;
}

static bool pcm_sysfs_write(const char *buf, unsigned long count)
{
	char field[30];
	int value;

	char desc[32];
	int len = 0;

	len = (count < (sizeof(desc) - 1)) ? count : (sizeof(desc) - 1);

	if (copy_from_user(desc, buf, len)) {
		return 0;
	}
	desc[len] = '\0';

	if (desc[0] == '0' && desc[1] == 'x')
		sscanf(desc, "%x%s", &value, field);
	else
		sscanf(desc, "%d%s", &value, field);

	spm_crit2("field = %s, value=%d\n", field, value);

	if (strcmp(field, "cpu_pdn") == 0) {
		if (pcm_fs_curr->scenario == SPM_PCM_KERNEL_SUSPEND)
			pcm_fs_curr->cpu_pdn = value;
		else if (pcm_fs_curr->scenario == SPM_PCM_DEEP_IDLE) {
			if (value == 0)
				idle_state_en(IDLE_TYPE_DP, value);	/* dpidle disable */
			else {
				pcm_fs_curr->cpu_pdn = (value - 1);
				idle_state_en(IDLE_TYPE_DP, 1);	/* dpidle enable */
			}
		}

	} else if (strcmp(field, "infra_pdn") == 0)
		pcm_fs_curr->infra_pdn = value;
	else if (strcmp(field, "timer_val_ms") == 0)
		pcm_fs_curr->timer_val_ms = value;
	else if (strcmp(field, "wfi_sel0") == 0)
		pcm_fs_curr->wfi_sel[0] = value;
	else if (strcmp(field, "wfi_sel1") == 0)
		pcm_fs_curr->wfi_sel[1] = value;
	else if (strcmp(field, "wake_src") == 0) {
		if (value == 0) {
			BUG_ON(pcm_wakesrc_backup[pcm_fs_curr->scenario] != 0);
			pcm_wakesrc_backup[pcm_fs_curr->scenario] =
			    spm_wakesrc_mask_all(pcm_fs_curr->scenario);
		} else if (value >= 0 && value < 32)
			spm_wakesrc_set(pcm_fs_curr->scenario, value);
		else if (value < 0 && value > -32)
			spm_wakesrc_clear(pcm_fs_curr->scenario, value * -1);
		else if (pcm_wakesrc_backup[pcm_fs_curr->scenario] != 0) {
			spm_wakesrc_restore(pcm_fs_curr->scenario,
					    pcm_wakesrc_backup[pcm_fs_curr->scenario]);
			pcm_wakesrc_backup[pcm_fs_curr->scenario] = 0;
		}

	} else if (strcmp(field, "md_mask") == 0) {
		if (value >= 0 && value <= 3)
			pcm_fs_curr->md_mask = value;
	} else if (strcmp(field, "reserved") == 0)
		pcm_fs_curr->reserved = value;
	else {
		return false;	/* Can't not find field */
	}
	return true;
}

#if 0
static bool pcm_sysfs_write(const char *buf, unsigned long count)
{
	char field[30];
	int value;

	if (buf[0] == '0' && buf[1] == 'x')
		sscanf(buf, "%x%s", &value, field);
	else
		sscanf(buf, "%d%s", &value, field);

	spm_crit2("field = %s, value=%d\n", field, value);

	if (strcmp(field, "cpu_pdn") == 0) {
		if (pcm_fs_curr->scenario == SPM_PCM_KERNEL_SUSPEND)
			pcm_fs_curr->cpu_pdn = value;
		else if (pcm_fs_curr->scenario == SPM_PCM_DEEP_IDLE) {
			if (value == 0)
				idle_state_en(IDLE_TYPE_DP, value);	/* dpidle disable */
			else {
				pcm_fs_curr->cpu_pdn = (value - 1);
				idle_state_en(IDLE_TYPE_DP, 1);	/* dpidle enable */
			}
		}

	} else if (strcmp(field, "infra_pdn") == 0)
		pcm_fs_curr->infra_pdn = value;
	else if (strcmp(field, "timer_val_ms") == 0)
		pcm_fs_curr->timer_val_ms = value;
	else if (strcmp(field, "wfi_sel0") == 0)
		pcm_fs_curr->wfi_sel[0] = value;
	else if (strcmp(field, "wfi_sel1") == 0)
		pcm_fs_curr->wfi_sel[1] = value;
	else if (strcmp(field, "wake_src") == 0) {
		if (value == 0) {
			BUG_ON(pcm_wakesrc_backup[pcm_fs_curr->scenario] != 0);
			pcm_wakesrc_backup[pcm_fs_curr->scenario] =
			    spm_wakesrc_mask_all(pcm_fs_curr->scenario);
		} else if (value >= 0 && value < 32)
			spm_wakesrc_set(pcm_fs_curr->scenario, value);
		else if (value < 0 && value > -32)
			spm_wakesrc_clear(pcm_fs_curr->scenario, value * -1);
		else if (pcm_wakesrc_backup[pcm_fs_curr->scenario] != 0) {
			spm_wakesrc_restore(pcm_fs_curr->scenario,
					    pcm_wakesrc_backup[pcm_fs_curr->scenario]);
			pcm_wakesrc_backup[pcm_fs_curr->scenario] = 0;
		}

	} else if (strcmp(field, "md_mask") == 0) {
		if (value >= 0 && value <= 3)
			pcm_fs_curr->md_mask = value;
	} else if (strcmp(field, "reserved") == 0)
		pcm_fs_curr->reserved = value;
	else {
		return false;	/* Can't not find field */
	}
	return true;
}

static int pcm_sysfs_read(char *buf)
{
	int len = 0;
	char *p = buf;

	if (pcm_fs_curr->scenario == SPM_PCM_KERNEL_SUSPEND)
		p += sprintf(p, "pcm_config_suspend={\n");
	else if (pcm_fs_curr->scenario == SPM_PCM_DEEP_IDLE)
		p += sprintf(p, "pcm_config_dpidle={\n");
	else if (pcm_fs_curr->scenario == SPM_PCM_MCDI)
		p += sprintf(p, "pcm_config_mcdi={\n");
	else {
		p += sprintf(p, "SCENARIO ERROR\n");
		goto error;
	}

	p += sprintf(p, "     .scenario = %d\n", pcm_fs_curr->scenario);
	p += sprintf(p, "     .spm_turn_off_26m = %d\n", pcm_fs_curr->spm_turn_off_26m);
	p += sprintf(p, "     .pcm_pwrlevel = %d\n", pcm_fs_curr->pcm_pwrlevel);
	p += sprintf(p, "     .pcm_firmware_addr = 0x%X\n", pcm_fs_curr->pcm_firmware_addr);
	p += sprintf(p, "     .pcm_firmware_len = %d\n", pcm_fs_curr->pcm_firmware_len);
	p += sprintf(p, "     .spm_request_uart_sleep = %d\n", pcm_fs_curr->spm_request_uart_sleep);
	/* p += sprintf(p,"     .pcm_vsr[SPM_PCM_MAX_VSR_NUM]; */
	p += sprintf(p,
		     "     .md_mask = 0x%X  /*{MDCONN_MASK=0,CONN_MASK,MD_MASK,MDCONN_UNMASK}*/\n",
		     pcm_fs_curr->md_mask);
	p += sprintf(p,
		     "     .mm_mask = 0x%X  /*{MMALL_MASK=0,MFG_MASK,DISP_MASK,MMALL_UNMASK}*/\n",
		     pcm_fs_curr->mm_mask);
	p += sprintf(p, "     .wfi_scu_mask = %d\n ", pcm_fs_curr->wfi_scu_mask);
	p += sprintf(p, "     .wfi_l2c_mask = %d\n ", pcm_fs_curr->wfi_l2c_mask);
	p += sprintf(p, "     .wfi_op = %d    /*{REDUCE_OR=0,REDUCE_AND}*/\n", pcm_fs_curr->wfi_op);
	p += sprintf(p, "     .wfi_sel[]={%d,%d}\n", pcm_fs_curr->wfi_sel[0],
		     pcm_fs_curr->wfi_sel[1]);
	p += sprintf(p, "     .timer_val_ms=%d\n", (u32) pcm_fs_curr->timer_val_ms);
	p += sprintf(p, "     .wake_src= 0x%X\n", pcm_fs_curr->wake_src);
	p += sprintf(p, "     .cpu_pdn= %d\n", pcm_fs_curr->cpu_pdn);
	p += sprintf(p, "     .infra_pdn = %d\n", pcm_fs_curr->infra_pdn);
	p += sprintf(p, "     .reserved = 0x%X\n", pcm_fs_curr->reserved);

	p += sprintf(p, "}\n");
error:
	len = p - buf;
	return len;
}
#endif

/*************************************************Suspend by Scott***************************************************************/


/* Sysfs Implementation  */
static int pcm_sysfs_read_suspend(struct seq_file *m, void *v)
{
	pcm_fs_curr = &pcm_config_suspend;
	pcm_sysfs_read(m, v);

	return 0;
}

static ssize_t pcm_sysfs_write_suspend(struct file *file, const char __user *buffer, size_t count,
				       loff_t *data)
{
	char field[30];
	int value;

	pcm_fs_curr = &pcm_config_suspend;

	if (!pcm_sysfs_write(buffer, count)) {
		/* No common field found in common, Try specific ones!! */
		if (buffer[0] == '0' && buffer[1] == 'x')
			sscanf(buffer, "%x%s", &value, field);
		else
			sscanf(buffer, "%d%s", &value, field);

		if (strcmp(field, "fgauge") == 0) {
			pcm_fs_curr->reserved &= (~(1U));
			pcm_fs_curr->reserved |= value;
		}
	}

	return count;
}

static int pcm_sysfs_read_suspend_mode(struct seq_file *m, void *v)
{
	pcm_fs_curr = &pcm_config_suspend;

	if (pcm_fs_curr->infra_pdn == 0 && pcm_fs_curr->infra_pdn == 0)
		seq_puts(m, "0");
	else if (pcm_fs_curr->infra_pdn == 1 && pcm_fs_curr->infra_pdn == 1)
		seq_puts(m, "1");
	else {
		BUG();
	}


	return 0;
}

static int pcm_sysfs_read_suspend_timer(struct seq_file *m, void *v)
{
	pcm_fs_curr = &pcm_config_suspend;

	if (pcm_fs_curr->reserved & SPM_SUSPEND_GET_FGUAGE)
		seq_puts(m, "1");
	else
		seq_puts(m, "0");

	seq_printf(m, "%d", (u32) pcm_fs_curr->timer_val_ms);

	return 0;
}

#if 0
static ssize_t pcm_sysfs_write_suspend(struct file *file, const char *buffer, unsigned long count,
				       void *data)
{
	char field[30];
	int value;

	pcm_fs_curr = &pcm_config_suspend;

	if (!pcm_sysfs_write(buffer, count)) {
		/* No common field found in common, Try specific ones!! */
		if (buffer[0] == '0' && buffer[1] == 'x')
			sscanf(buffer, "%x%s", &value, field);
		else
			sscanf(buffer, "%d%s", &value, field);

		if (strcmp(field, "fgauge") == 0) {
			pcm_fs_curr->reserved &= (~(1U));
			pcm_fs_curr->reserved |= value;
		}
	}


	return count;
}

static int pcm_sysfs_read_suspend(char *buf, char **start, off_t off, int count, int *eof,
				  void *data)
{
	pcm_fs_curr = &pcm_config_suspend;

	return pcm_sysfs_read(buf);
}

static int pcm_sysfs_read_suspend_mode(char *buf, char **start, off_t off, int count, int *eof,
				       void *data)
{
	int len = 0;
	char *p = buf;
	pcm_fs_curr = &pcm_config_suspend;

	if (pcm_fs_curr->infra_pdn == 0 && pcm_fs_curr->infra_pdn == 0)
		p += sprintf(p, "0");
	else if (pcm_fs_curr->infra_pdn == 1 && pcm_fs_curr->infra_pdn == 1)
		p += sprintf(p, "1");
	else {
		BUG();
	}

	len = p - buf;

	return len;
}

static int pcm_sysfs_read_suspend_timer(char *buf, char **start, off_t off, int count, int *eof,
					void *data)
{
	int len = 0;
	char *p = buf;
	pcm_fs_curr = &pcm_config_suspend;

	if (pcm_fs_curr->reserved & SPM_SUSPEND_GET_FGUAGE)
		p += sprintf(p, "1 ");
	else
		p += sprintf(p, "0 ");

	p += sprintf(p, "%d", (u32) pcm_fs_curr->timer_val_ms);

	len = p - buf;

	return len;
}
#endif

/*************************************************dpidle & mcdi by Mark***************************************************************/

static int pcm_sysfs_read_dpidle(struct seq_file *m, void *v)
{
	pcm_fs_curr = &pcm_config_dpidle;

	return pcm_sysfs_read(m, v);
}

static ssize_t pcm_sysfs_write_dpidle(struct file *file, const char __user *buffer, size_t count,
				      loff_t *data)
{

	char field[30];
	int value;

	pcm_fs_curr = &pcm_config_dpidle;


	if (!pcm_sysfs_write(buffer, count)) {
		/* No common field found in common, Try specific ones!! */
		if (buffer[0] == '0' || buffer[0] == 'x')
			sscanf(buffer, "%x%s", &value, field);
		else
			sscanf(buffer, "%d%s", &value, field);

		if (strcmp(field, "pwrlevel") == 0)
			pcm_fs_curr->pcm_pwrlevel = (1 << value);

	}

	return count;
}

static int pcm_sysfs_read_dpidle_mode(struct seq_file *m, void *v)
{
	pcm_fs_curr = &pcm_config_dpidle;

	if (idle_state_get(IDLE_TYPE_DP) == 0)
		seq_puts(m, "0");
	else if (pcm_fs_curr->cpu_pdn == 0)	/* Legacy Mode */
		seq_puts(m, "1");
	else if (pcm_fs_curr->cpu_pdn == 1)	/* CPU dormant */
		seq_puts(m, "2");
	else {
		BUG();
	}

	return 0;
}

static int pcm_sysfs_read_dpidle_level(struct seq_file *m, void *v)
{
	pcm_fs_curr = &pcm_config_dpidle;

	if (pcm_fs_curr->pcm_pwrlevel == PWR_LV0)
		seq_puts(m, "0");
	else if (pcm_fs_curr->pcm_pwrlevel == PWR_LV1)
		seq_puts(m, "1");
	else {
		BUG();
	}

	return 0;
}

static int pcm_sysfs_read_dpidle_timer(struct seq_file *m, void *v)
{
	pcm_fs_curr = &pcm_config_dpidle;

	seq_printf(m, "%d", (u32) pcm_fs_curr->timer_val_ms);

	return 0;
}

#if 0
static ssize_t pcm_sysfs_write_dpidle(struct file *file, const char *buffer, unsigned long count,
				      void *data)
{

	char field[30];
	int value;

	pcm_fs_curr = &pcm_config_dpidle;


	if (!pcm_sysfs_write(buffer, count)) {
		/* No common field found in common, Try specific ones!! */
		if (buffer[0] == '0' || buffer[0] == 'x')
			sscanf(buffer, "%x%s", &value, field);
		else
			sscanf(buffer, "%d%s", &value, field);

		if (strcmp(field, "pwrlevel") == 0)
			pcm_fs_curr->pcm_pwrlevel = (1 << value);

	}

	return count;
}

static int pcm_sysfs_read_dpidle(char *buf, char **start, off_t off, int count, int *eof,
				 void *data)
{
	pcm_fs_curr = &pcm_config_dpidle;

	return pcm_sysfs_read(buf);
}

static int pcm_sysfs_read_dpidle_mode(char *buf, char **start, off_t off, int count, int *eof,
				      void *data)
{
	int len = 0;
	char *p = buf;
	pcm_fs_curr = &pcm_config_dpidle;

	if (idle_state_get(IDLE_TYPE_DP) == 0)
		p += sprintf(p, "0");
	else if (pcm_fs_curr->cpu_pdn == 0)	/* Legacy Mode */
		p += sprintf(p, "1");
	else if (pcm_fs_curr->cpu_pdn == 1)	/* CPU dormant */
		p += sprintf(p, "2");
	else {
		BUG();
	}

	len = p - buf;

	return len;
}

static int pcm_sysfs_read_dpidle_level(char *buf, char **start, off_t off, int count, int *eof,
				       void *data)
{
	int len = 0;
	char *p = buf;
	pcm_fs_curr = &pcm_config_dpidle;

	if (pcm_fs_curr->pcm_pwrlevel == PWR_LV0)
		p += sprintf(p, "0");
	else if (pcm_fs_curr->pcm_pwrlevel == PWR_LV1)
		p += sprintf(p, "1");
	else {
		BUG();
	}

	len = p - buf;

	return len;
}

static int pcm_sysfs_read_dpidle_timer(char *buf, char **start, off_t off, int count, int *eof,
				       void *data)
{
	int len = 0;
	char *p = buf;
	pcm_fs_curr = &pcm_config_dpidle;

	p += sprintf(p, "%d", (u32) pcm_fs_curr->timer_val_ms);

	len = p - buf;

	return len;
}
#endif

static int pcm_sysfs_read_mcdi(struct seq_file *m, void *v)
{
	pcm_fs_curr = &pcm_config_mcdi;

	pcm_sysfs_read(m, v);

	return pcm_sysfs_read(m, v);
}

static ssize_t pcm_sysfs_write_mcdi(struct file *file, const char __user *buffer, size_t count,
				    loff_t *data)
{
	char field[30];
	int value;

	pcm_fs_curr = &pcm_config_mcdi;


	if (!pcm_sysfs_write(buffer, count)) {
		/* No common field found in common, Try specific ones!! */
		if (buffer[0] == '0' || buffer[0] == 'x')
			sscanf(buffer, "%x%s", &value, field);
		else
			sscanf(buffer, "%d%s", &value, field);

		if (strcmp(field, "mcdi_mode") == 0) {
			if (value == 0)
				idle_state_en(IDLE_TYPE_MC, value);	/* Disable NCDI */
			else {
				if (value == 1)
					spm_disable_sodi_user();
				if (value == 2)
					spm_enable_sodi_user();
				idle_state_en(IDLE_TYPE_MC, 1);	/* MCDI handler */
			}
		}


	}


	return count;
}

static int pcm_sysfs_read_mcdi_mode(struct seq_file *m, void *v)
{
	pcm_fs_curr = &pcm_config_mcdi;

	if (idle_state_get(IDLE_TYPE_MC) == 0)
		seq_puts(m, "0");
	else if (spm_is_sodi_user_en() == false)	/* MCDI Only */
		seq_puts(m, "1");
	else
		seq_puts(m, "2");

	return 0;
}

static int pcm_sysfs_read_mcdi_timer(struct seq_file *m, void *v)
{
	pcm_fs_curr = &pcm_config_mcdi;

	seq_printf(m, "%d", (u32) pcm_fs_curr->timer_val_ms);

	return 0;
}

#if 0
static ssize_t pcm_sysfs_write_mcdi(struct file *file, const char *buffer, unsigned long count,
				    void *data)
{
	char field[30];
	int value;

	pcm_fs_curr = &pcm_config_mcdi;


	if (!pcm_sysfs_write(buffer, count)) {
		/* No common field found in common, Try specific ones!! */
		if (buffer[0] == '0' || buffer[0] == 'x')
			sscanf(buffer, "%x%s", &value, field);
		else
			sscanf(buffer, "%d%s", &value, field);

		if (strcmp(field, "mcdi_mode") == 0) {
			if (value == 0)
				idle_state_en(IDLE_TYPE_MC, value);	/* Disable NCDI */
			else {
				if (value == 1)
					spm_disable_sodi_user();
				if (value == 2)
					spm_enable_sodi_user();
				idle_state_en(IDLE_TYPE_MC, 1);	/* MCDI handler */
			}
		}


	}


	return count;
}

static int pcm_sysfs_read_mcdi(char *buf, char **start, off_t off, int count, int *eof, void *data)
{
	pcm_fs_curr = &pcm_config_mcdi;

	pcm_sysfs_read(buf);

	return pcm_sysfs_read(buf);
}

static int pcm_sysfs_read_mcdi_mode(char *buf, char **start, off_t off, int count, int *eof,
				    void *data)
{
	int len = 0;
	char *p = buf;
	pcm_fs_curr = &pcm_config_mcdi;

	if (idle_state_get(IDLE_TYPE_MC) == 0)
		p += sprintf(p, "0");
	else if (spm_is_sodi_user_en() == false)	/* MCDI Only */
		p += sprintf(p, "1");
	else
		p += sprintf(p, "2");

	len = p - buf;

	return len;
}

static int pcm_sysfs_read_mcdi_timer(char *buf, char **start, off_t off, int count, int *eof,
				     void *data)
{
	int len = 0;
	char *p = buf;
	pcm_fs_curr = &pcm_config_mcdi;

	p += sprintf(p, "%d", (u32) pcm_fs_curr->timer_val_ms);

	len = p - buf;

	return len;
}
#endif

static int pcm_sysfs_open_mcdi_timer(struct inode *inode, struct file *file)
{
	return single_open(file, pcm_sysfs_read_mcdi_timer, NULL);
}

static const struct file_operations pcm_sysfs_mcdi_timer_fops = {
	.owner = THIS_MODULE,
	.open = pcm_sysfs_open_mcdi_timer,
	.read = seq_read,
};

static int pcm_sysfs_open_mcdi_mode(struct inode *inode, struct file *file)
{
	return single_open(file, pcm_sysfs_read_mcdi_mode, NULL);
}

static const struct file_operations pcm_sysfs_mcdi_mode_fops = {
	.owner = THIS_MODULE,
	.open = pcm_sysfs_open_mcdi_mode,
	.read = seq_read,
};

static int pcm_sysfs_open_mcdi(struct inode *inode, struct file *file)
{
	return single_open(file, pcm_sysfs_read_mcdi, NULL);
}

static const struct file_operations pcm_sysfs_mcdi_fops = {
	.owner = THIS_MODULE,
	.open = pcm_sysfs_open_mcdi,
	.read = seq_read,
	.llseek = seq_lseek,
	.write = pcm_sysfs_write_mcdi,
	.release = single_release,
};

static int pcm_sysfs_open_dpidle_timer(struct inode *inode, struct file *file)
{
	return single_open(file, pcm_sysfs_read_dpidle_timer, NULL);
}

static const struct file_operations pcm_sysfs_dpidle_timer_fops = {
	.owner = THIS_MODULE,
	.open = pcm_sysfs_open_dpidle_timer,
	.read = seq_read,
};

static int pcm_sysfs_open_dpidle_level(struct inode *inode, struct file *file)
{
	return single_open(file, pcm_sysfs_read_dpidle_level, NULL);
}

static const struct file_operations pcm_sysfs_dpidle_level_fops = {
	.owner = THIS_MODULE,
	.open = pcm_sysfs_open_dpidle_level,
	.read = seq_read,
};

static int pcm_sysfs_open_dpidle_mode(struct inode *inode, struct file *file)
{
	return single_open(file, pcm_sysfs_read_dpidle_mode, NULL);
}

static const struct file_operations pcm_sysfs_dpidle_mode_fops = {
	.owner = THIS_MODULE,
	.open = pcm_sysfs_open_dpidle_mode,
	.read = seq_read,
};

static int pcm_sysfs_open_dpidle(struct inode *inode, struct file *file)
{
	return single_open(file, pcm_sysfs_read_dpidle, NULL);
}

static const struct file_operations pcm_sysfs_dpidle_fops = {
	.owner = THIS_MODULE,
	.open = pcm_sysfs_open_dpidle,
	.read = seq_read,
	.llseek = seq_lseek,
	.write = pcm_sysfs_write_dpidle,
	.release = single_release,
};

static int pcm_sysfs_open_suspend_timer(struct inode *inode, struct file *file)
{
	return single_open(file, pcm_sysfs_read_suspend_timer, NULL);
}

static const struct file_operations pcm_sysfs_suspend_timer_fops = {
	.owner = THIS_MODULE,
	.open = pcm_sysfs_open_suspend_timer,
	.read = seq_read,
};

static int pcm_sysfs_open_suspend_mode(struct inode *inode, struct file *file)
{
	return single_open(file, pcm_sysfs_read_suspend_mode, NULL);
}

static const struct file_operations pcm_sysfs_suspend_mode_fops = {
	.owner = THIS_MODULE,
	.open = pcm_sysfs_open_suspend_mode,
	.read = seq_read,
};

static int pcm_sysfs_open_suspend(struct inode *inode, struct file *file)
{
	return single_open(file, pcm_sysfs_read_suspend, NULL);
}

static const struct file_operations pcm_sysfs_suspend_fops = {
	.owner = THIS_MODULE,
	.open = pcm_sysfs_open_suspend,
	.read = seq_read,
	.llseek = seq_lseek,
	.write = pcm_sysfs_write_suspend,
	.release = single_release,
};

void spm_fs_init(void)
{
	struct proc_dir_entry *spm_fs_file = NULL;
	struct proc_dir_entry *spm_fs_dir = NULL;

	spm_fs_dir = proc_mkdir("spm_fs", NULL);
	if (!spm_fs_dir) {
		/* clc_notice("[%s]: mkdir /proc/mcdi failed\n", __FUNCTION__); */
	} else {
 /****************************************Suspend**********************************************************/
		spm_fs_file =
		    proc_create("suspend", S_IRUGO | S_IWUSR | S_IWGRP, spm_fs_dir,
				&pcm_sysfs_suspend_fops);
		spm_fs_file =
		    proc_create("suspend_mode", S_IRUGO | S_IWUSR | S_IWGRP, spm_fs_dir,
				&pcm_sysfs_suspend_mode_fops);
		spm_fs_file =
		    proc_create("suspend_timer", S_IRUGO | S_IWUSR | S_IWGRP, spm_fs_dir,
				&pcm_sysfs_suspend_timer_fops);

#if 0
		spm_fs_file = create_proc_entry("suspend", S_IRUGO | S_IWUSR | S_IWGRP, spm_fs_dir);
		if (spm_fs_file) {
			spm_fs_file->read_proc = pcm_sysfs_read_suspend;
			spm_fs_file->write_proc = pcm_sysfs_write_suspend;
		}

		spm_fs_file =
		    create_proc_entry("suspend_mode", S_IRUGO | S_IWUSR | S_IWGRP, spm_fs_dir);
		if (spm_fs_file) {
			spm_fs_file->read_proc = pcm_sysfs_read_suspend_mode;
		}
		spm_fs_file =
		    create_proc_entry("suspend_timer", S_IRUGO | S_IWUSR | S_IWGRP, spm_fs_dir);
		if (spm_fs_file) {
			spm_fs_file->read_proc = pcm_sysfs_read_suspend_timer;
		}
#endif
  /****************************************Dpidle**********************************************************/
		spm_fs_file =
		    proc_create("dpidle", S_IRUGO | S_IWUSR | S_IWGRP, spm_fs_dir,
				&pcm_sysfs_dpidle_fops);
		spm_fs_file =
		    proc_create("dpidle_mode", S_IRUGO | S_IWUSR | S_IWGRP, spm_fs_dir,
				&pcm_sysfs_dpidle_mode_fops);
		spm_fs_file =
		    proc_create("dpidle_level", S_IRUGO | S_IWUSR | S_IWGRP, spm_fs_dir,
				&pcm_sysfs_dpidle_level_fops);
		spm_fs_file =
		    proc_create("dpidle_timer", S_IRUGO | S_IWUSR | S_IWGRP, spm_fs_dir,
				&pcm_sysfs_dpidle_timer_fops);

#if 0
		spm_fs_file = create_proc_entry("dpidle", S_IRUGO | S_IWUSR | S_IWGRP, spm_fs_dir);
		if (spm_fs_file) {
			spm_fs_file->read_proc = pcm_sysfs_read_dpidle;
			spm_fs_file->write_proc = pcm_sysfs_write_dpidle;
		}
		spm_fs_file =
		    create_proc_entry("dpidle_mode", S_IRUGO | S_IWUSR | S_IWGRP, spm_fs_dir);
		if (spm_fs_file) {
			spm_fs_file->read_proc = pcm_sysfs_read_dpidle_mode;
		}
		spm_fs_file =
		    create_proc_entry("dpidle_level", S_IRUGO | S_IWUSR | S_IWGRP, spm_fs_dir);
		if (spm_fs_file) {
			spm_fs_file->read_proc = pcm_sysfs_read_dpidle_level;
		}
		spm_fs_file =
		    create_proc_entry("dpidle_timer", S_IRUGO | S_IWUSR | S_IWGRP, spm_fs_dir);
		if (spm_fs_file) {
			spm_fs_file->read_proc = pcm_sysfs_read_dpidle_timer;
		}
#endif
  /****************************************MCDI**********************************************************/
		spm_fs_file =
		    proc_create("mcdi", S_IRUGO | S_IWUSR | S_IWGRP, spm_fs_dir,
				&pcm_sysfs_mcdi_fops);
		spm_fs_file =
		    proc_create("mcdi_mode", S_IRUGO | S_IWUSR | S_IWGRP, spm_fs_dir,
				&pcm_sysfs_mcdi_mode_fops);
		spm_fs_file =
		    proc_create("mcdi_timer", S_IRUGO | S_IWUSR | S_IWGRP, spm_fs_dir,
				&pcm_sysfs_mcdi_timer_fops);

#if 0
		spm_fs_file = create_proc_entry("mcdi", S_IRUGO | S_IWUSR | S_IWGRP, spm_fs_dir);
		if (spm_fs_file) {
			spm_fs_file->read_proc = pcm_sysfs_read_mcdi;
			spm_fs_file->write_proc = pcm_sysfs_write_mcdi;
		}
		spm_fs_file =
		    create_proc_entry("mcdi_mode", S_IRUGO | S_IWUSR | S_IWGRP, spm_fs_dir);
		if (spm_fs_file) {
			spm_fs_file->read_proc = pcm_sysfs_read_mcdi_mode;
		}
		spm_fs_file =
		    create_proc_entry("mcdi_timer", S_IRUGO | S_IWUSR | S_IWGRP, spm_fs_dir);
		if (spm_fs_file) {
			spm_fs_file->read_proc = pcm_sysfs_read_mcdi_timer;
		}
#endif

	}

}
