/*
* Copyright (C) 2011-2014 MediaTek Inc.
*
* This program is free software: you can redistribute it and/or modify it under the terms of the
* GNU General Public License version 2 as published by the Free Software Foundation.
*
* This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
* without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
* See the GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License along with this program.
* If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef __DDP_REG_H__
#define __DDP_REG_H__

#include <mach/mt_reg_base.h>
#include <mt-plat/sync_write.h>

#define DISPSYS_COLOR_BASE      DISP_COLOR_BASE
#define DISPSYS_CMDQ_BASE       MM_CMDQ_BASE

#define DISPSYS_REG_ADDR_MIN    MMSYS_CONFIG_BASE
#define DISPSYS_REG_ADDR_MAX    (DISP_DBI_BASE+0x2000)
/* --------------------------------------------------------------------------- */
/* Register Field Access */
/* --------------------------------------------------------------------------- */

#define REG_FLD(width, shift) \
	((unsigned int)((((width) & 0xFF) << 16) | ((shift) & 0xFF)))

#define REG_FLD_WIDTH(field) \
	((unsigned int)(((field) >> 16) & 0xFF))

#define REG_FLD_SHIFT(field) \
	((unsigned int)((field) & 0xFF))

#define REG_FLD_MASK(field) \
	(((unsigned int)(1 << REG_FLD_WIDTH(field)) - 1) << REG_FLD_SHIFT(field))

#define REG_FLD_VAL(field, val) \
	(((val) << REG_FLD_SHIFT(field)) & REG_FLD_MASK(field))

#define REG_FLD_GET(field, reg32) \
	(((reg32) & REG_FLD_MASK(field)) >> REG_FLD_SHIFT(field))

#define REG_FLD_SET(field, reg32, val) \
	do { \
		(reg32) = (((reg32) & ~REG_FLD_MASK(field)) | \
		REG_FLD_VAL((field), (val))); \
	} while (0)

#define DISP_REG_GET(reg32) (*(volatile unsigned int*)(reg32))
#define DISP_REG_GET_FIELD(field, reg32) ((*(volatile unsigned int*)(reg32) & REG_FLD_MASK(field)) >> REG_FLD_SHIFT(field))

#define DISP_REG_SET(reg32, val) mt_reg_sync_writel(val, (volatile unsigned int *)(reg32))
void DISP_REG_SET_FIELD(unsigned long field, unsigned long reg32, unsigned long val);

#define DDP_MOD_MDP_WROT    BIT(0)
#define DDP_MOD_MDP_PRZ0    BIT(1)
#define DDP_MOD_MDP_PRZ1    BIT(2)
#define DDP_MOD_DISP_OVL    BIT(3)
#define DDP_MOD_MDP_WDMA    BIT(4)
#define DDP_MOD_MDP_RDMA    BIT(5)
#define DDP_MOD_DISP_WDMA   BIT(6)
#define DDP_MOD_DISP_COLOR  BIT(7)
#define DDP_MOD_MDP_TDSHP   BIT(8)
#define DDP_MOD_DISP_BLS    BIT(9)
#define DDP_MOD_DISP_RDMA   BIT(10)
#define DDP_MOD_ISP_MOUT    BIT(11)

/* ----------------------------------------------------------------- */
/* CMDQ */
#define DISP_REG_CMDQ_IRQ_FLAG                           (DISPSYS_CMDQ_BASE + 0x10)
#define DISP_REG_CMDQ_LOADED_THR                         (DISPSYS_CMDQ_BASE + 0x20)
#define DISP_REG_CMDQ_THR_SLOT_CYCLES                    (DISPSYS_CMDQ_BASE + 0x30)
#define DISP_REG_CMDQ_BUS_CTRL                           (DISPSYS_CMDQ_BASE + 0x40)
#define DISP_REG_CMDQ_ABORT                              (DISPSYS_CMDQ_BASE + 0x50)
#define DISP_REG_CMDQ_SYNC_TOKEN_UPDATE                  (DISPSYS_CMDQ_BASE + 0x68)
#define DISP_REG_CMDQ_THRx_RESET(idx)                    (DISPSYS_CMDQ_BASE + 0x100 + 0x80*idx)
#define DISP_REG_CMDQ_THRx_EN(idx)                       (DISPSYS_CMDQ_BASE + 0x104 + 0x80*idx)
#define DISP_REG_CMDQ_THRx_SUSPEND(idx)                  (DISPSYS_CMDQ_BASE + 0x108 + 0x80*idx)
#define DISP_REG_CMDQ_THRx_STATUS(idx)                   (DISPSYS_CMDQ_BASE + 0x10c + 0x80*idx)
#define DISP_REG_CMDQ_THRx_IRQ_FLAG(idx)                 (DISPSYS_CMDQ_BASE + 0x110 + 0x80*idx)
#define DISP_REG_CMDQ_THRx_IRQ_FLAG_EN(idx)              (DISPSYS_CMDQ_BASE + 0x114 + 0x80*idx)
#define DISP_REG_CMDQ_THRx_SECURITY(idx)                 (DISPSYS_CMDQ_BASE + 0x118 + 0x80*idx)
#define DISP_REG_CMDQ_THRx_PC(idx)                       (DISPSYS_CMDQ_BASE + 0x120 + 0x80*idx)
#define DISP_REG_CMDQ_THRx_END_ADDR(idx)                 (DISPSYS_CMDQ_BASE + 0x124 + 0x80*idx)
#define DISP_REG_CMDQ_THRx_EXEC_CMDS_CNT(idx)            (DISPSYS_CMDQ_BASE + 0x128 + 0x80*idx)
#define DISP_REG_CMDQ_THRx_WAIT_EVENTS0(idx)             (DISPSYS_CMDQ_BASE + 0x130 + 0x80*idx)
#define DISP_REG_CMDQ_THRx_WAIT_EVENTS1(idx)             (DISPSYS_CMDQ_BASE + 0x134 + 0x80*idx)
#define DISP_REG_CMDQ_THRx_OBSERVED_EVENTS0(idx)         (DISPSYS_CMDQ_BASE + 0x140 + 0x80*idx)
#define DISP_REG_CMDQ_THRx_OBSERVED_EVENTS1(idx)         (DISPSYS_CMDQ_BASE + 0x144 + 0x80*idx)
#define DISP_REG_CMDQ_THRx_OBSERVED_EVENTS0_CLR(idx)     (DISPSYS_CMDQ_BASE + 0x148 + 0x80*idx)
#define DISP_REG_CMDQ_THRx_OBSERVED_EVENTS1_CLR(idx)     (DISPSYS_CMDQ_BASE + 0x14c + 0x80*idx)
#define DISP_REG_CMDQ_THRx_INSTN_TIMEOUT_CYCLES(idx)     (DISPSYS_CMDQ_BASE + 0x150 + 0x80*idx)

/* ----------------------------------------------------------------- */
/* Config */
#define DISP_REG_CONFIG_RDMA_MOUT_EN              (MMSYS_CONFIG_BASE + 0x020)
#define DISP_REG_CONFIG_RSZ0_MOUT_EN              (MMSYS_CONFIG_BASE + 0x024)
#define DISP_REG_CONFIG_RSZ1_MOUT_EN              (MMSYS_CONFIG_BASE + 0x028)
#define DISP_REG_CONFIG_TDSHP_MOUT_EN             (MMSYS_CONFIG_BASE + 0x02C)
#define DISP_REG_CONFIG_DISP_OVL_MOUT_EN          (MMSYS_CONFIG_BASE + 0x030)
#define DISP_REG_CONFIG_MOUT_RST                  (MMSYS_CONFIG_BASE + 0x034)
#define DISP_REG_CONFIG_RSZ0_SEL                  (MMSYS_CONFIG_BASE + 0x038)
#define DISP_REG_CONFIG_RSZ1_SEL                  (MMSYS_CONFIG_BASE + 0x03C)
#define DISP_REG_CONFIG_TDSHP_SEL                 (MMSYS_CONFIG_BASE + 0x040)
#define DISP_REG_CONFIG_WROT_SEL                  (MMSYS_CONFIG_BASE + 0x044)
#define DISP_REG_CONFIG_WDMA0_SEL                 (MMSYS_CONFIG_BASE + 0x048)
#define DISP_REG_CONFIG_DISP_OUT_SEL              (MMSYS_CONFIG_BASE + 0x04C)
#define DISP_REG_CONFIG_RDMA0_OUT_SEL             (MMSYS_CONFIG_BASE + 0x050)
#define DISP_REG_CONFIG_COLOR_SEL                 (MMSYS_CONFIG_BASE + 0x054)
#define DISP_REG_CONFIG_DSI_SEL                   (MMSYS_CONFIG_BASE + 0x058)
#define DISP_REG_CONFIG_DPI0_SEL                  (MMSYS_CONFIG_BASE + 0x05C)
#define DISP_REG_CONFIG_MMSYS_CG_CON0             (MMSYS_CONFIG_BASE + 0x100)
#define DISP_REG_CONFIG_MMSYS_CG_SET0             (MMSYS_CONFIG_BASE + 0x104)
#define DISP_REG_CONFIG_MMSYS_CG_CLR0             (MMSYS_CONFIG_BASE + 0x108)
#define DISP_REG_CONFIG_MMSYS_CG_CON1             (MMSYS_CONFIG_BASE + 0x110)
#define DISP_REG_CONFIG_MMSYS_CG_SET1             (MMSYS_CONFIG_BASE + 0x114)
#define DISP_REG_CONFIG_MMSYS_CG_CLR1             (MMSYS_CONFIG_BASE + 0x118)
#define DISP_REG_CONFIG_HW_DCM_EN0                (MMSYS_CONFIG_BASE + 0x120)
#define DISP_REG_CONFIG_HW_DCM_EN_SET0            (MMSYS_CONFIG_BASE + 0x124)
#define DISP_REG_CONFIG_HW_DCM_EN_CLR0            (MMSYS_CONFIG_BASE + 0x128)
#define DISP_REG_CONFIG_HW_DCM_EN1                (MMSYS_CONFIG_BASE + 0x12C)
#define DISP_REG_CONFIG_HW_DCM_EN_SET1            (MMSYS_CONFIG_BASE + 0x130)
#define DISP_REG_CONFIG_HW_DCM_EN_CLR1            (MMSYS_CONFIG_BASE + 0x134)
#define DISP_REG_CONFIG_MMSYS_SW_RST_B            (MMSYS_CONFIG_BASE + 0x138)
#define DISP_REG_CONFIG_MBIST_DONE0               (MMSYS_CONFIG_BASE + 0x800)
#define DISP_REG_CONFIG_MBIST_FAIL0               (MMSYS_CONFIG_BASE + 0x804)
#define DISP_REG_CONFIG_MBIST_FAIL1               (MMSYS_CONFIG_BASE + 0x808)
#define DISP_REG_CONFIG_MBIST_HOLDB0              (MMSYS_CONFIG_BASE + 0x80C)
#define DISP_REG_CONFIG_MBIST_MODE0               (MMSYS_CONFIG_BASE + 0x810)
#define DISP_REG_CONFIG_MBIST_BSEL0               (MMSYS_CONFIG_BASE + 0x814)
#define DISP_REG_CONFIG_MBIST_BSEL1               (MMSYS_CONFIG_BASE + 0x818)
#define DISP_REG_CONFIG_MBIST_BSEL2               (MMSYS_CONFIG_BASE + 0x81C)
#define DISP_REG_CONFIG_MBIST_CON                 (MMSYS_CONFIG_BASE + 0x820)
#define DISP_REG_CONFIG_MEM_DELSEL0               (MMSYS_CONFIG_BASE + 0x824)
#define DISP_REG_CONFIG_MEM_DELSEL1               (MMSYS_CONFIG_BASE + 0x828)
#define DISP_REG_CONFIG_MEM_DELSEL2               (MMSYS_CONFIG_BASE + 0x82C)
#define DISP_REG_CONFIG_MEM_DELSEL3               (MMSYS_CONFIG_BASE + 0x830)
#define DISP_REG_CONFIG_MEM_DELSEL4               (MMSYS_CONFIG_BASE + 0x834)
#define DISP_REG_CONFIG_MEM_DELSEL5               (MMSYS_CONFIG_BASE + 0x838)
#define DISP_REG_CONFIG_MEM_DELSEL6               (MMSYS_CONFIG_BASE + 0x83C)
#define DISP_REG_CONFIG_DEBUG_OUT_SEL             (MMSYS_CONFIG_BASE + 0x840)
#define DISP_REG_CONFIG_DUMMY                     (MMSYS_CONFIG_BASE + 0x844)
#define DISP_REG_CONFIG_WROT_MBISR_RESET          (MMSYS_CONFIG_BASE + 0x850)
#define DISP_REG_CONFIG_WROT_MBISR_FAIL           (MMSYS_CONFIG_BASE + 0x854)
#define DISP_REG_CONFIG_WROT_MBISR_OK             (MMSYS_CONFIG_BASE + 0x858)
#define DISP_REG_CONFIG_DL_VALID0                 (MMSYS_CONFIG_BASE + 0x860)
#define DISP_REG_CONFIG_DL_VALID1                 (MMSYS_CONFIG_BASE + 0x864)
#define DISP_REG_CONFIG_DL_READY0                 (MMSYS_CONFIG_BASE + 0x868)
#define DISP_REG_CONFIG_DL_READY1                 (MMSYS_CONFIG_BASE + 0x86C)

/* ----------------------------------------------------------------- */
/* Mutex */
#define DISP_REG_CONFIG_MUTEX_INTEN               (MM_MUTEX_BASE + 0x000)
#define DISP_REG_CONFIG_MUTEX_INTSTA              (MM_MUTEX_BASE + 0x004)
#define DISP_REG_CONFIG_REG_UPD_TIMEOUT           (MM_MUTEX_BASE + 0x008)
#define DISP_REG_CONFIG_REG_COMMIT                (MM_MUTEX_BASE + 0x00C)
#define DISP_REG_CONFIG_MUTEX_EN(n)               (MM_MUTEX_BASE + 0x020 + (0x20 * n))
#define DISP_REG_CONFIG_MUTEX(n)                  (MM_MUTEX_BASE + 0x024 + (0x20 * n))
#define DISP_REG_CONFIG_MUTEX_RST(n)              (MM_MUTEX_BASE + 0x028 + (0x20 * n))
#define DISP_REG_CONFIG_MUTEX_MOD(n)              (MM_MUTEX_BASE + 0x02C + (0x20 * n))
#define DISP_REG_CONFIG_MUTEX_SOF(n)              (MM_MUTEX_BASE + 0x030 + (0x20 * n))
#define DISP_REG_CONFIG_MUTEX0_EN                 (MM_MUTEX_BASE + 0x020)
#define DISP_REG_CONFIG_MUTEX0                    (MM_MUTEX_BASE + 0x024)
#define DISP_REG_CONFIG_MUTEX0_RST                (MM_MUTEX_BASE + 0x028)
#define DISP_REG_CONFIG_MUTEX0_MOD                (MM_MUTEX_BASE + 0x02C)
#define DISP_REG_CONFIG_MUTEX0_SOF                (MM_MUTEX_BASE + 0x030)
#define DISP_REG_CONFIG_MUTEX1_EN                 (MM_MUTEX_BASE + 0x040)
#define DISP_REG_CONFIG_MUTEX1                    (MM_MUTEX_BASE + 0x044)
#define DISP_REG_CONFIG_MUTEX1_RST                (MM_MUTEX_BASE + 0x048)
#define DISP_REG_CONFIG_MUTEX1_MOD                (MM_MUTEX_BASE + 0x04C)
#define DISP_REG_CONFIG_MUTEX1_SOF                (MM_MUTEX_BASE + 0x050)
#define DISP_REG_CONFIG_MUTEX2_EN                 (MM_MUTEX_BASE + 0x060)
#define DISP_REG_CONFIG_MUTEX2                    (MM_MUTEX_BASE + 0x064)
#define DISP_REG_CONFIG_MUTEX2_RST                (MM_MUTEX_BASE + 0x068)
#define DISP_REG_CONFIG_MUTEX2_MOD                (MM_MUTEX_BASE + 0x06C)
#define DISP_REG_CONFIG_MUTEX2_SOF                (MM_MUTEX_BASE + 0x070)
#define DISP_REG_CONFIG_MUTEX3_EN                 (MM_MUTEX_BASE + 0x080)
#define DISP_REG_CONFIG_MUTEX3                    (MM_MUTEX_BASE + 0x084)
#define DISP_REG_CONFIG_MUTEX3_RST                (MM_MUTEX_BASE + 0x088)
#define DISP_REG_CONFIG_MUTEX3_MOD                (MM_MUTEX_BASE + 0x08C)
#define DISP_REG_CONFIG_MUTEX3_SOF                (MM_MUTEX_BASE + 0x090)
#define DISP_REG_CONFIG_MUTEX4_EN                 (MM_MUTEX_BASE + 0x0A0)
#define DISP_REG_CONFIG_MUTEX4                    (MM_MUTEX_BASE + 0x0A4)
#define DISP_REG_CONFIG_MUTEX4_RST                (MM_MUTEX_BASE + 0x0A8)
#define DISP_REG_CONFIG_MUTEX4_MOD                (MM_MUTEX_BASE + 0x0AC)
#define DISP_REG_CONFIG_MUTEX4_SOF                (MM_MUTEX_BASE + 0x0B0)
#define DISP_REG_CONFIG_MUTEX5_EN                 (MM_MUTEX_BASE + 0x0C0)
#define DISP_REG_CONFIG_MUTEX5                    (MM_MUTEX_BASE + 0x0C4)
#define DISP_REG_CONFIG_MUTEX5_RST                (MM_MUTEX_BASE + 0x0C8)
#define DISP_REG_CONFIG_MUTEX5_MOD                (MM_MUTEX_BASE + 0x0CC)
#define DISP_REG_CONFIG_MUTEX5_SOF                (MM_MUTEX_BASE + 0x0D0)
#define DISP_REG_CONFIG_MUTEX6_EN                 (MM_MUTEX_BASE + 0x0E0)
#define DISP_REG_CONFIG_MUTEX6                    (MM_MUTEX_BASE + 0x0E4)
#define DISP_REG_CONFIG_MUTEX6_RST                (MM_MUTEX_BASE + 0x0E8)
#define DISP_REG_CONFIG_MUTEX6_MOD                (MM_MUTEX_BASE + 0x0EC)
#define DISP_REG_CONFIG_MUTEX6_SOF                (MM_MUTEX_BASE + 0x0F0)
#define DISP_REG_CONFIG_MUTEX7_EN                 (MM_MUTEX_BASE + 0x100)
#define DISP_REG_CONFIG_MUTEX7                    (MM_MUTEX_BASE + 0x104)
#define DISP_REG_CONFIG_MUTEX7_RST                (MM_MUTEX_BASE + 0x108)
#define DISP_REG_CONFIG_MUTEX7_MOD                (MM_MUTEX_BASE + 0x10C)
#define DISP_REG_CONFIG_MUTEX7_SOF                (MM_MUTEX_BASE + 0x110)
#define DISP_REG_CONFIG_MUTEX_DEBUG_OUT_SEL       (MM_MUTEX_BASE + 0x200)

/* ----------------------------------------------------------------- */
/* OVL */
#define DISP_REG_OVL_STA                          (DISP_OVL_BASE + 0x000)
#define DISP_REG_OVL_INTEN                        (DISP_OVL_BASE + 0x004)
#define DISP_REG_OVL_INTSTA                       (DISP_OVL_BASE + 0x008)
#define DISP_REG_OVL_EN                           (DISP_OVL_BASE + 0x00C)
#define DISP_REG_OVL_TRIG                         (DISP_OVL_BASE + 0x010)
#define DISP_REG_OVL_RST                          (DISP_OVL_BASE + 0x014)
#define DISP_REG_OVL_ROI_SIZE                     (DISP_OVL_BASE + 0x020)
#define DISP_REG_OVL_DATAPATH_CON                 (DISP_OVL_BASE + 0x024)
#define DISP_REG_OVL_ROI_BGCLR                    (DISP_OVL_BASE + 0x028)
#define DISP_REG_OVL_SRC_CON                      (DISP_OVL_BASE + 0x02C)
#define DISP_REG_OVL_L0_CON                       (DISP_OVL_BASE + 0x030)
#define DISP_REG_OVL_L0_SRCKEY                    (DISP_OVL_BASE + 0x034)
#define DISP_REG_OVL_L0_SRC_SIZE                  (DISP_OVL_BASE + 0x038)
#define DISP_REG_OVL_L0_OFFSET                    (DISP_OVL_BASE + 0x03C)
#define DISP_REG_OVL_L0_ADDR                      (DISP_OVL_BASE + 0x040)
#define DISP_REG_OVL_L0_PITCH                     (DISP_OVL_BASE + 0x044)
#define DISP_REG_OVL_L1_CON                       (DISP_OVL_BASE + 0x050)
#define DISP_REG_OVL_L1_SRCKEY                    (DISP_OVL_BASE + 0x054)
#define DISP_REG_OVL_L1_SRC_SIZE                  (DISP_OVL_BASE + 0x058)
#define DISP_REG_OVL_L1_OFFSET                    (DISP_OVL_BASE + 0x05C)
#define DISP_REG_OVL_L1_ADDR                      (DISP_OVL_BASE + 0x060)
#define DISP_REG_OVL_L1_PITCH                     (DISP_OVL_BASE + 0x064)
#define DISP_REG_OVL_L2_CON                       (DISP_OVL_BASE + 0x070)
#define DISP_REG_OVL_L2_SRCKEY                    (DISP_OVL_BASE + 0x074)
#define DISP_REG_OVL_L2_SRC_SIZE                  (DISP_OVL_BASE + 0x078)
#define DISP_REG_OVL_L2_OFFSET                    (DISP_OVL_BASE + 0x07C)
#define DISP_REG_OVL_L2_ADDR                      (DISP_OVL_BASE + 0x080)
#define DISP_REG_OVL_L2_PITCH                     (DISP_OVL_BASE + 0x084)
#define DISP_REG_OVL_L3_CON                       (DISP_OVL_BASE + 0x090)
#define DISP_REG_OVL_L3_SRCKEY                    (DISP_OVL_BASE + 0x094)
#define DISP_REG_OVL_L3_SRC_SIZE                  (DISP_OVL_BASE + 0x098)
#define DISP_REG_OVL_L3_OFFSET                    (DISP_OVL_BASE + 0x09C)
#define DISP_REG_OVL_L3_ADDR                      (DISP_OVL_BASE + 0x0A0)
#define DISP_REG_OVL_L3_PITCH                     (DISP_OVL_BASE + 0x0A4)
#define DISP_REG_OVL_RDMA0_CTRL                   (DISP_OVL_BASE + 0x0C0)
#define DISP_REG_OVL_RDMA0_MEM_START_TRIG         (DISP_OVL_BASE + 0x0C4)
#define DISP_REG_OVL_RDMA0_MEM_GMC_SETTING        (DISP_OVL_BASE + 0x0C8)
#define DISP_REG_OVL_RDMA0_MEM_SLOW_CON           (DISP_OVL_BASE + 0x0CC)
#define DISP_REG_OVL_RDMA0_FIFO_CTRL              (DISP_OVL_BASE + 0x0D0)
#define DISP_REG_OVL_RDMA1_CTRL                   (DISP_OVL_BASE + 0x0E0)
#define DISP_REG_OVL_RDMA1_MEM_START_TRIG         (DISP_OVL_BASE + 0x0E4)
#define DISP_REG_OVL_RDMA1_MEM_GMC_SETTING        (DISP_OVL_BASE + 0x0E8)
#define DISP_REG_OVL_RDMA1_MEM_SLOW_CON           (DISP_OVL_BASE + 0x0EC)
#define DISP_REG_OVL_RDMA1_FIFO_CTRL              (DISP_OVL_BASE + 0x0F0)
#define DISP_REG_OVL_RDMA2_CTRL                   (DISP_OVL_BASE + 0x100)
#define DISP_REG_OVL_RDMA2_MEM_START_TRIG         (DISP_OVL_BASE + 0x104)
#define DISP_REG_OVL_RDMA2_MEM_GMC_SETTING        (DISP_OVL_BASE + 0x108)
#define DISP_REG_OVL_RDMA2_MEM_SLOW_CON           (DISP_OVL_BASE + 0x10C)
#define DISP_REG_OVL_RDMA2_FIFO_CTRL              (DISP_OVL_BASE + 0x110)
#define DISP_REG_OVL_RDMA3_CTRL                   (DISP_OVL_BASE + 0x120)
#define DISP_REG_OVL_RDMA3_MEM_START_TRIG         (DISP_OVL_BASE + 0x124)
#define DISP_REG_OVL_RDMA3_MEM_GMC_SETTING        (DISP_OVL_BASE + 0x128)
#define DISP_REG_OVL_RDMA3_MEM_SLOW_CON           (DISP_OVL_BASE + 0x12C)
#define DISP_REG_OVL_RDMA3_FIFO_CTRL              (DISP_OVL_BASE + 0x130)
#define DISP_REG_OVL_L0_Y2R_PARA_R0               (DISP_OVL_BASE + 0x134)
#define DISP_REG_OVL_L0_Y2R_PARA_R1               (DISP_OVL_BASE + 0x138)
#define DISP_REG_OVL_L0_Y2R_PARA_G0               (DISP_OVL_BASE + 0x13C)
#define DISP_REG_OVL_L0_Y2R_PARA_G1               (DISP_OVL_BASE + 0x140)
#define DISP_REG_OVL_L0_Y2R_PARA_B0               (DISP_OVL_BASE + 0x144)
#define DISP_REG_OVL_L0_Y2R_PARA_B1               (DISP_OVL_BASE + 0x148)
#define DISP_REG_OVL_L0_Y2R_PARA_YUV_A_0          (DISP_OVL_BASE + 0x14C)
#define DISP_REG_OVL_L0_Y2R_PARA_YUV_A_1          (DISP_OVL_BASE + 0x150)
#define DISP_REG_OVL_L0_Y2R_PARA_RGB_A_0          (DISP_OVL_BASE + 0x154)
#define DISP_REG_OVL_L0_Y2R_PARA_RGB_A_1          (DISP_OVL_BASE + 0x158)
#define DISP_REG_OVL_L1_Y2R_PARA_R0               (DISP_OVL_BASE + 0x15C)
#define DISP_REG_OVL_L1_Y2R_PARA_R1               (DISP_OVL_BASE + 0x160)
#define DISP_REG_OVL_L1_Y2R_PARA_G0               (DISP_OVL_BASE + 0x164)
#define DISP_REG_OVL_L1_Y2R_PARA_G1               (DISP_OVL_BASE + 0x168)
#define DISP_REG_OVL_L1_Y2R_PARA_B0               (DISP_OVL_BASE + 0x16C)
#define DISP_REG_OVL_L1_Y2R_PARA_B1               (DISP_OVL_BASE + 0x170)
#define DISP_REG_OVL_L1_Y2R_PARA_YUV_A_0          (DISP_OVL_BASE + 0x174)
#define DISP_REG_OVL_L1_Y2R_PARA_YUV_A_1          (DISP_OVL_BASE + 0x178)
#define DISP_REG_OVL_L1_Y2R_PARA_RGB_A_0          (DISP_OVL_BASE + 0x17C)
#define DISP_REG_OVL_L1_Y2R_PARA_RGB_A_1          (DISP_OVL_BASE + 0x180)
#define DISP_REG_OVL_L2_Y2R_PARA_R0               (DISP_OVL_BASE + 0x184)
#define DISP_REG_OVL_L2_Y2R_PARA_R1               (DISP_OVL_BASE + 0x188)
#define DISP_REG_OVL_L2_Y2R_PARA_G0               (DISP_OVL_BASE + 0x18C)
#define DISP_REG_OVL_L2_Y2R_PARA_G1               (DISP_OVL_BASE + 0x190)
#define DISP_REG_OVL_L2_Y2R_PARA_B0               (DISP_OVL_BASE + 0x194)
#define DISP_REG_OVL_L2_Y2R_PARA_B1               (DISP_OVL_BASE + 0x198)
#define DISP_REG_OVL_L2_Y2R_PARA_YUV_A_0          (DISP_OVL_BASE + 0x19C)
#define DISP_REG_OVL_L2_Y2R_PARA_YUV_A_1          (DISP_OVL_BASE + 0x1A0)
#define DISP_REG_OVL_L2_Y2R_PARA_RGB_A_0          (DISP_OVL_BASE + 0x1A4)
#define DISP_REG_OVL_L2_Y2R_PARA_RGB_A_1          (DISP_OVL_BASE + 0x1A8)
#define DISP_REG_OVL_L3_Y2R_PARA_R0               (DISP_OVL_BASE + 0x1AC)
#define DISP_REG_OVL_L3_Y2R_PARA_R1               (DISP_OVL_BASE + 0x1B0)
#define DISP_REG_OVL_L3_Y2R_PARA_G0               (DISP_OVL_BASE + 0x1B4)
#define DISP_REG_OVL_L3_Y2R_PARA_G1               (DISP_OVL_BASE + 0x1B8)
#define DISP_REG_OVL_L3_Y2R_PARA_B0               (DISP_OVL_BASE + 0x1BC)
#define DISP_REG_OVL_L3_Y2R_PARA_B1               (DISP_OVL_BASE + 0x1C0)
#define DISP_REG_OVL_L3_Y2R_PARA_YUV_A_0          (DISP_OVL_BASE + 0x1C4)
#define DISP_REG_OVL_L3_Y2R_PARA_YUV_A_1          (DISP_OVL_BASE + 0x1C8)
#define DISP_REG_OVL_L3_Y2R_PARA_RGB_A_0          (DISP_OVL_BASE + 0x1CC)
#define DISP_REG_OVL_L3_Y2R_PARA_RGB_A_1          (DISP_OVL_BASE + 0x1D0)
#define DISP_REG_OVL_DEBUG_MON_SEL                (DISP_OVL_BASE + 0x1D4)
#define DISP_REG_OVL_RDMA0_MEM_GMC_SETTING2       (DISP_OVL_BASE + 0x1E0)
#define DISP_REG_OVL_RDMA1_MEM_GMC_SETTING2       (DISP_OVL_BASE + 0x1E4)
#define DISP_REG_OVL_RDMA2_MEM_GMC_SETTING2       (DISP_OVL_BASE + 0x1E8)
#define DISP_REG_OVL_RDMA3_MEM_GMC_SETTING2       (DISP_OVL_BASE + 0x1EC)
#define DISP_REG_OVL_DUMMY                        (DISP_OVL_BASE + 0x200)
#define DISP_REG_OVL_FLOW_CTRL_DBG                (DISP_OVL_BASE + 0x240)
#define DISP_REG_OVL_ADDCON_DBG                   (DISP_OVL_BASE + 0x244)
#define DISP_REG_OVL_OUTMUX_DBG                   (DISP_OVL_BASE + 0x248)
#define DISP_REG_OVL_RDMA0_DBG                    (DISP_OVL_BASE + 0x24C)
#define DISP_REG_OVL_RDMA1_DBG                    (DISP_OVL_BASE + 0x250)
#define DISP_REG_OVL_RDMA2_DBG                    (DISP_OVL_BASE + 0x254)
#define DISP_REG_OVL_RDMA3_DBG                    (DISP_OVL_BASE + 0x258)

/* ----------------------------------------------------------------- */
/* RDMA */
#define DISP_REG_RDMA_INT_ENABLE                  (DISP_RDMA_BASE + 0x000)
#define DISP_REG_RDMA_INT_STATUS                  (DISP_RDMA_BASE + 0x004)
#define DISP_REG_RDMA_GLOBAL_CON                  (DISP_RDMA_BASE + 0x010)
#define DISP_REG_RDMA_SIZE_CON_0                  (DISP_RDMA_BASE + 0x014)
#define DISP_REG_RDMA_SIZE_CON_1                  (DISP_RDMA_BASE + 0x018)
#define DISP_REG_RDMA_TARGET_LINE                 (DISP_RDMA_BASE + 0x01C)
#define DISP_REG_RDMA_MEM_CON                     (DISP_RDMA_BASE + 0x024)
#define DISP_REG_RDMA_MEM_START_ADDR              (DISP_RDMA_BASE + 0x028)
#define DISP_REG_RDMA_MEM_SRC_PITCH               (DISP_RDMA_BASE + 0x02C)
#define DISP_REG_RDMA_MEM_GMC_SETTING_0           (DISP_RDMA_BASE + 0x030)
#define DISP_REG_RDMA_MEM_SLOW_CON                (DISP_RDMA_BASE + 0x034)
#define DISP_REG_RDMA_MEM_GMC_SETTING_1           (DISP_RDMA_BASE + 0x038)
#define DISP_REG_RDMA_FIFO_CON                    (DISP_RDMA_BASE + 0x040)
#define DISP_REG_RDMA_FIFO_LOG                    (DISP_RDMA_BASE + 0x044)
#define DISP_REG_RDMA_CF_00                       (DISP_RDMA_BASE + 0x054)
#define DISP_REG_RDMA_CF_01                       (DISP_RDMA_BASE + 0x058)
#define DISP_REG_RDMA_CF_02                       (DISP_RDMA_BASE + 0x05C)
#define DISP_REG_RDMA_CF_10                       (DISP_RDMA_BASE + 0x060)
#define DISP_REG_RDMA_CF_11                       (DISP_RDMA_BASE + 0x064)
#define DISP_REG_RDMA_CF_12                       (DISP_RDMA_BASE + 0x068)
#define DISP_REG_RDMA_CF_20                       (DISP_RDMA_BASE + 0x06C)
#define DISP_REG_RDMA_CF_21                       (DISP_RDMA_BASE + 0x070)
#define DISP_REG_RDMA_CF_22                       (DISP_RDMA_BASE + 0x074)
#define DISP_REG_RDMA_CF_PRE_ADD0                 (DISP_RDMA_BASE + 0x078)
#define DISP_REG_RDMA_CF_PRE_ADD1                 (DISP_RDMA_BASE + 0x07C)
#define DISP_REG_RDMA_CF_PRE_ADD2                 (DISP_RDMA_BASE + 0x080)
#define DISP_REG_RDMA_CF_POST_ADD0                (DISP_RDMA_BASE + 0x084)
#define DISP_REG_RDMA_CF_POST_ADD1                (DISP_RDMA_BASE + 0x088)
#define DISP_REG_RDMA_CF_POST_ADD2                (DISP_RDMA_BASE + 0x08C)
#define DISP_REG_RDMA_DUMMY                       (DISP_RDMA_BASE + 0x090)
#define DISP_REG_RDMA_DEBUG_OUT_SEL               (DISP_RDMA_BASE + 0x094)

/* ----------------------------------------------------------------- */
/* WDMA */
#define DISP_REG_WDMA_INTEN                       (DISP_WDMA_BASE + 0x000)
#define DISP_REG_WDMA_INTSTA                      (DISP_WDMA_BASE + 0x004)
#define DISP_REG_WDMA_EN                          (DISP_WDMA_BASE + 0x008)
#define DISP_REG_WDMA_RST                         (DISP_WDMA_BASE + 0x00C)
#define DISP_REG_WDMA_SMI_CON                     (DISP_WDMA_BASE + 0x010)
#define DISP_REG_WDMA_CFG                         (DISP_WDMA_BASE + 0x014)
#define DISP_REG_WDMA_SRC_SIZE                    (DISP_WDMA_BASE + 0x018)
#define DISP_REG_WDMA_CLIP_SIZE                   (DISP_WDMA_BASE + 0x01C)
#define DISP_REG_WDMA_CLIP_COORD                  (DISP_WDMA_BASE + 0x020)
#define DISP_REG_WDMA_DST_ADDR                    (DISP_WDMA_BASE + 0x024)
#define DISP_REG_WDMA_DST_W_IN_BYTE               (DISP_WDMA_BASE + 0x028)
#define DISP_REG_WDMA_ALPHA                       (DISP_WDMA_BASE + 0x02C)
#define DISP_REG_WDMA_BUF_ADDR                    (DISP_WDMA_BASE + 0x030)
#define DISP_REG_WDMA_STA                         (DISP_WDMA_BASE + 0x034)
#define DISP_REG_WDMA_BUF_CON1                    (DISP_WDMA_BASE + 0x038)
#define DISP_REG_WDMA_BUF_CON2                    (DISP_WDMA_BASE + 0x03C)
#define DISP_REG_WDMA_C00                         (DISP_WDMA_BASE + 0x040)
#define DISP_REG_WDMA_C02                         (DISP_WDMA_BASE + 0x044)
#define DISP_REG_WDMA_C10                         (DISP_WDMA_BASE + 0x048)
#define DISP_REG_WDMA_C12                         (DISP_WDMA_BASE + 0x04C)
#define DISP_REG_WDMA_C20                         (DISP_WDMA_BASE + 0x050)
#define DISP_REG_WDMA_C22                         (DISP_WDMA_BASE + 0x054)
#define DISP_REG_WDMA_PRE_ADD0                    (DISP_WDMA_BASE + 0x058)
#define DISP_REG_WDMA_PRE_ADD2                    (DISP_WDMA_BASE + 0x05C)
#define DISP_REG_WDMA_POST_ADD0                   (DISP_WDMA_BASE + 0x060)
#define DISP_REG_WDMA_POST_ADD2                   (DISP_WDMA_BASE + 0x064)
#define DISP_REG_WDMA_DST_U_ADDR                  (DISP_WDMA_BASE + 0x070)
#define DISP_REG_WDMA_DST_V_ADDR                  (DISP_WDMA_BASE + 0x074)
#define DISP_REG_WDMA_DST_UV_PITCH                (DISP_WDMA_BASE + 0x078)
#define DISP_REG_WDMA_DITHER_CON                  (DISP_WDMA_BASE + 0x090)
#define DISP_REG_WDMA_FLOW_CTRL_DBG               (DISP_WDMA_BASE + 0x0A0)
#define DISP_REG_WDMA_EXEC_DBG                    (DISP_WDMA_BASE + 0x0A4)
#define DISP_REG_WDMA_CLIP_DBG                    (DISP_WDMA_BASE + 0x0A8)

/* ----------------------------------------------------------------- */
/* COLOR */
#define DISP_REG_COLOR_CFG_MAIN                   (DISP_COLOR_BASE + 0x400)
#define DISP_REG_COLOR_PXL_CNT_MAIN               (DISP_COLOR_BASE + 0x404)
#define DISP_REG_COLOR_LINE_CNT_MAIN              (DISP_COLOR_BASE + 0x408)
#define DISP_REG_COLOR_WIN_X_MAIN                 (DISP_COLOR_BASE + 0x40C)
#define DISP_REG_COLOR_WIN_Y_MAIN                 (DISP_COLOR_BASE + 0x410)
#define DISP_REG_COLOR_TIMING_DETECTION_0         (DISP_COLOR_BASE + 0x418)
#define DISP_REG_COLOR_TIMING_DETECTION_1         (DISP_COLOR_BASE + 0x41C)
#define DISP_REG_COLOR_DBG_CFG_MAIN               (DISP_COLOR_BASE + 0x420)
#define DISP_REG_COLOR_C_BOOST_MAIN               (DISP_COLOR_BASE + 0x428)
#define DISP_REG_COLOR_C_BOOST_MAIN_2             (DISP_COLOR_BASE + 0x42C)
#define DISP_REG_COLOR_LUMA_ADJ                   (DISP_COLOR_BASE + 0x430)
#define DISP_REG_COLOR_G_PIC_ADJ_MAIN_1           (DISP_COLOR_BASE + 0x434)
#define DISP_REG_COLOR_G_PIC_ADJ_MAIN_2           (DISP_COLOR_BASE + 0x438)
#define DISP_REG_COLOR_Y_FTN_1_0_MAIN             (DISP_COLOR_BASE + 0x440)
#define DISP_REG_COLOR_Y_FTN_3_2_MAIN             (DISP_COLOR_BASE + 0x444)
#define DISP_REG_COLOR_Y_FTN_5_4_MAIN             (DISP_COLOR_BASE + 0x448)
#define DISP_REG_COLOR_Y_FTN_7_6_MAIN             (DISP_COLOR_BASE + 0x44C)
#define DISP_REG_COLOR_Y_FTN_9_8_MAIN             (DISP_COLOR_BASE + 0x450)
#define DISP_REG_COLOR_Y_FTN_11_10_MAIN           (DISP_COLOR_BASE + 0x454)
#define DISP_REG_COLOR_Y_FTN_13_12_MAIN           (DISP_COLOR_BASE + 0x458)
#define DISP_REG_COLOR_Y_FTN_15_14_MAIN           (DISP_COLOR_BASE + 0x45C)
#define DISP_REG_COLOR_Y_FTN_17_16_MAIN           (DISP_COLOR_BASE + 0x460)
#define DISP_REG_COLOR_POS_MAIN                   (DISP_COLOR_BASE + 0x484)
#define DISP_REG_COLOR_INK_DATA_MAIN              (DISP_COLOR_BASE + 0x488)
#define DISP_REG_COLOR_INK_DATA_MAIN_CR           (DISP_COLOR_BASE + 0x48C)
#define DISP_REG_COLOR_CAP_IN_DATA_MAIN           (DISP_COLOR_BASE + 0x490)
#define DISP_REG_COLOR_CAP_IN_DATA_MAIN_CR        (DISP_COLOR_BASE + 0x494)
#define DISP_REG_COLOR_CAP_OUT_DATA_MAIN          (DISP_COLOR_BASE + 0x498)
#define DISP_REG_COLOR_CAP_OUT_DATA_MAIN_CR       (DISP_COLOR_BASE + 0x49C)
#define DISP_REG_COLOR_LUMA_HIST_00               (DISP_COLOR_BASE + 0x520)
#define DISP_REG_COLOR_LUMA_HIST_01               (DISP_COLOR_BASE + 0x524)
#define DISP_REG_COLOR_LUMA_HIST_02               (DISP_COLOR_BASE + 0x528)
#define DISP_REG_COLOR_LUMA_HIST_03               (DISP_COLOR_BASE + 0x52C)
#define DISP_REG_COLOR_LUMA_HIST_04               (DISP_COLOR_BASE + 0x530)
#define DISP_REG_COLOR_LUMA_HIST_05               (DISP_COLOR_BASE + 0x534)
#define DISP_REG_COLOR_LUMA_HIST_06               (DISP_COLOR_BASE + 0x538)
#define DISP_REG_COLOR_LUMA_HIST_07               (DISP_COLOR_BASE + 0x53C)
#define DISP_REG_COLOR_LUMA_HIST_08               (DISP_COLOR_BASE + 0x540)
#define DISP_REG_COLOR_LUMA_HIST_09               (DISP_COLOR_BASE + 0x544)
#define DISP_REG_COLOR_LUMA_HIST_10               (DISP_COLOR_BASE + 0x548)
#define DISP_REG_COLOR_LUMA_HIST_11               (DISP_COLOR_BASE + 0x54C)
#define DISP_REG_COLOR_LUMA_HIST_12               (DISP_COLOR_BASE + 0x550)
#define DISP_REG_COLOR_LUMA_HIST_13               (DISP_COLOR_BASE + 0x554)
#define DISP_REG_COLOR_LUMA_HIST_14               (DISP_COLOR_BASE + 0x558)
#define DISP_REG_COLOR_LUMA_HIST_15               (DISP_COLOR_BASE + 0x55C)
#define DISP_REG_COLOR_LUMA_HIST_16               (DISP_COLOR_BASE + 0x560)
#define DISP_REG_COLOR_LUMA_SUM                   (DISP_COLOR_BASE + 0x5A4)
#define DISP_REG_COLOR_LUMA_MIN_MAX               (DISP_COLOR_BASE + 0x5A8)
#define DISP_REG_COLOR_LOCAL_HUE_CD_0             (DISP_COLOR_BASE + 0x620)
#define DISP_REG_COLOR_LOCAL_HUE_CD_1             (DISP_COLOR_BASE + 0x624)
#define DISP_REG_COLOR_LOCAL_HUE_CD_2             (DISP_COLOR_BASE + 0x628)
#define DISP_REG_COLOR_LOCAL_HUE_CD_3             (DISP_COLOR_BASE + 0x62C)
#define DISP_REG_COLOR_LOCAL_HUE_CD_4             (DISP_COLOR_BASE + 0x630)
#define DISP_REG_COLOR_TWO_D_WINDOW_1             (DISP_COLOR_BASE + 0x740)
#define DISP_REG_COLOR_TWO_D_W1_RESULT            (DISP_COLOR_BASE + 0x74C)
#define DISP_REG_COLOR_SAT_HIST_X_CFG_MAIN        (DISP_COLOR_BASE + 0x768)
#define DISP_REG_COLOR_SAT_HIST_Y_CFG_MAIN        (DISP_COLOR_BASE + 0x76C)
#define DISP_REG_COLOR_CRC(n)                     (DISP_COLOR_BASE + 0x7E0 + (4 * n))
#define DISP_REG_COLOR_PARTIAL_SAT_GAIN1(n)       (DISP_COLOR_BASE + 0x7FC + (4 * n))
#define DISP_REG_COLOR_PARTIAL_SAT_GAIN2(n)       (DISP_COLOR_BASE + 0x810 + (4 * n))
#define DISP_REG_COLOR_PARTIAL_SAT_GAIN3(n)       (DISP_COLOR_BASE + 0x824 + (4 * n))
#define DISP_REG_COLOR_PARTIAL_SAT_POINT1(n)      (DISP_COLOR_BASE + 0x838 + (4 * n))
#define DISP_REG_COLOR_PARTIAL_SAT_POINT2(n)      (DISP_COLOR_BASE + 0x84C + (4 * n))
#define DISP_REG_COLOR_START                      (DISP_COLOR_BASE + 0xF00)
#define DISP_REG_COLOR_INTEN                      (DISP_COLOR_BASE + 0xF04)
#define DISP_REG_COLOR_INTSTA                     (DISP_COLOR_BASE + 0xF08)
#define DISP_REG_COLOR_OUT_SEL                    (DISP_COLOR_BASE + 0xF0C)
#define DISP_REG_COLOR_FRAME_DONE_DEL             (DISP_COLOR_BASE + 0xF10)
#define DISP_REG_COLOR_CRC_EN                     (DISP_COLOR_BASE + 0xF14)
#define DISP_REG_COLOR_SW_SCRATCH                 (DISP_COLOR_BASE + 0xF18)
#define DISP_REG_COLOR_RDY_SEL                    (DISP_COLOR_BASE + 0xF20)
#define DISP_REG_COLOR_RDY_SEL_EN                 (DISP_COLOR_BASE + 0xF24)
#define DISP_REG_COLOR_CK_ON                      (DISP_COLOR_BASE + 0xF28)
#define DISP_REG_COLOR_INTERNAL_IP_WIDTH          (DISP_COLOR_BASE + 0xF50)
#define DISP_REG_COLOR_INTERNAL_IP_HEIGHT         (DISP_COLOR_BASE + 0xF54)
#define DISP_REG_COLOR_CM1_EN                     (DISP_COLOR_BASE + 0xF60)
#define DISP_REG_COLOR_CM2_EN                     (DISP_COLOR_BASE + 0xFA0)
#define DISP_REG_COLOR_R0_CRC                     (DISP_COLOR_BASE + 0xFF0)

/* ----------------------------------------------------------------- */
/* BLS */
#define DISP_REG_BLS_EN                           (DISP_BLS_BASE + 0x000)
#define DISP_REG_BLS_RST                          (DISP_BLS_BASE + 0x004)
#define DISP_REG_BLS_INTEN                        (DISP_BLS_BASE + 0x008)
#define DISP_REG_BLS_INTSTA                       (DISP_BLS_BASE + 0x00C)
#define DISP_REG_BLS_BLS_SETTING                  (DISP_BLS_BASE + 0x010)
#define DISP_REG_BLS_FANA_SETTING                 (DISP_BLS_BASE + 0x014)
#define DISP_REG_BLS_SRC_SIZE                     (DISP_BLS_BASE + 0x018)
#define DISP_REG_BLS_GAIN_SETTING                 (DISP_BLS_BASE + 0x020)
#define DISP_REG_BLS_MANUAL_GAIN                  (DISP_BLS_BASE + 0x024)
#define DISP_REG_BLS_MANUAL_MAXCLR                (DISP_BLS_BASE + 0x028)
#define DISP_REG_BLS_GAMMA_SETTING                (DISP_BLS_BASE + 0x030)
#define DISP_REG_BLS_GAMMA_BOUNDARY               (DISP_BLS_BASE + 0x034)
#define DISP_REG_BLS_LUT_UPDATE                   (DISP_BLS_BASE + 0x038)
#define DISP_REG_BLS_MAXCLR_THD                   (DISP_BLS_BASE + 0x060)
#define DISP_REG_BLS_DISTPT_THD                   (DISP_BLS_BASE + 0x064)
#define DISP_REG_BLS_MAXCLR_LIMIT                 (DISP_BLS_BASE + 0x068)
#define DISP_REG_BLS_DISTPT_LIMIT                 (DISP_BLS_BASE + 0x06C)
#define DISP_REG_BLS_AVE_SETTING                  (DISP_BLS_BASE + 0x070)
#define DISP_REG_BLS_AVE_LIMIT                    (DISP_BLS_BASE + 0x074)
#define DISP_REG_BLS_DISTPT_SETTING               (DISP_BLS_BASE + 0x078)
#define DISP_REG_BLS_HIS_CLEAR                    (DISP_BLS_BASE + 0x07C)
#define DISP_REG_BLS_SC_DIFF_THD                  (DISP_BLS_BASE + 0x080)
#define DISP_REG_BLS_SC_BIN_THD                   (DISP_BLS_BASE + 0x084)
#define DISP_REG_BLS_MAXCLR_GRADUAL               (DISP_BLS_BASE + 0x088)
#define DISP_REG_BLS_DISTPT_GRADUAL               (DISP_BLS_BASE + 0x08C)
#define DISP_REG_BLS_FAST_IIR_XCOEFF              (DISP_BLS_BASE + 0x090)
#define DISP_REG_BLS_FAST_IIR_YCOEFF              (DISP_BLS_BASE + 0x094)
#define DISP_REG_BLS_SLOW_IIR_XCOEFF              (DISP_BLS_BASE + 0x098)
#define DISP_REG_BLS_SLOW_IIR_YCOEFF              (DISP_BLS_BASE + 0x09C)
#define DISP_REG_BLS_PWM_DUTY                     (DISP_BLS_BASE + 0x0A0)
#define DISP_REG_BLS_PWM_GRADUAL                  (DISP_BLS_BASE + 0x0A4)
#define DISP_REG_BLS_PWM_CON                      (DISP_BLS_BASE + 0x0A8)
#define DISP_REG_BLS_PWM_MANUAL                   (DISP_BLS_BASE + 0x0AC)
#define DISP_REG_BLS_DEBUG                        (DISP_BLS_BASE + 0x0B0)
#define DISP_REG_BLS_PATTERN                      (DISP_BLS_BASE + 0x0B4)
#define DISP_REG_BLS_CHKSUM                       (DISP_BLS_BASE + 0x0B8)
#define DISP_REG_BLS_HIS_BIN(X)                   (DISP_BLS_BASE + 0x100 + (4*(X)))
#define DISP_REG_BLS_PWM_DUTY_RD                  (DISP_BLS_BASE + 0x200)
#define DISP_REG_BLS_FRAME_AVE_RD                 (DISP_BLS_BASE + 0x204)
#define DISP_REG_BLS_MAXCLR_RD                    (DISP_BLS_BASE + 0x208)
#define DISP_REG_BLS_DISTPT_RD                    (DISP_BLS_BASE + 0x20C)
#define DISP_REG_BLS_GAIN_RD                      (DISP_BLS_BASE + 0x210)
#define DISP_REG_BLS_SC_RD                        (DISP_BLS_BASE + 0x214)
#define DISP_REG_BLS_LUMINANCE(X)                 (DISP_BLS_BASE + 0x300 + (4*(X)))
#define DISP_REG_BLS_LUMINANCE_255                (DISP_BLS_BASE + 0x384)
#define DISP_REG_BLS_GAMMA_LUT(X)                 (DISP_BLS_BASE + 0x400 + (4*(X)))
#define DISP_REG_BLS_DITHER(X)                    (DISP_BLS_BASE + 0xE00 + (4*(X)))
#define DISP_REG_BLS_DUMMY                        (DISP_BLS_BASE + 0xF00)

#endif
